;;;  Name:  ACT-R Vision
;;;  Author:  Roman Belavkin
;;;  Contact: rvb@Cs.Nott.AC.UK
;;;  Version: 0.6c
;;;  Date: 17/07/2001
;;;  Description:  Visualisation tools for ACT-R
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  This code was written at the University of Nottingham, ;;;
;;;  and has been placed in the public domain.              ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Note: This program requires Garnet package and ACT-R
;;;------------------------------------------------------------

;;; If you are using CMU CL, then the two commands bellow may
;;; improve the performance

;(setf *gc-verbose* nil)
;(setf *bytes-consed-between-gcs* 4000000)

;;; Load Garnet if needed

;(load "home:garnet/src/garnet-loader.lisp") ;; CHANGE AND UNCOMMENT

;;; Load ACT-R

;(load "home:lisp/ACT-R4/ACT-R_4.0.lisp") ;; CHANGE AND UNCOMMENT

;;; Load necessary Garnet components

(defvar VISION-INIT
  (dolist (file '("motif-text-buttons-loader"
                  "motif-check-buttons-loader"
                  "motif-menubar-loader"
                  "motif-scrolling-labeled-box-loader"
                  "motif-error-gadget-loader"
                  "motif-save-gadget-loader"))
          (user::garnet-load (concatenate 'string "gadgets:" file))))

;;; Define globals

(defvar *vision-version* "0.6c")

(defvar *about-vision* (concatenate 'string
"ACT-R Visualisation Tools
Version " *vision-version*))

(defvar *vision-colour* opal:motif-light-green
  "This variable is initialised bellow to set the general colour")

(defvar *not-my-windows* nil
  "Used to destroy only Vision's windows")

(defvar *chunk-watch-list* nil
  "This list contains pairs of chunk names and their watch objects names")

(defvar *rule-watch-list* nil
  "This list contains pairs of rule names and their watch objects names")

(defvar *goal-stack-watch-p* nil)

(defvar *goal-focus-watch-p* nil)

;;; Initialise globals

(setf *vision-colour* opal:motif-light-green) ;; change the colour here
(setf *not-my-windows* (windows))
(setf *chunk-watch-list* nil)
(setf *rule-watch-list* nil)
(setf *goal-stack-watch-p* nil)
(setf *goal-focus-watch-p* nil)

;;; Useful functions

(defun unwrap (list)
  (and list (if (listp list) (unwrap (car list)) list)))

(defmacro watch-command (command text-object)
  "Redirects the output of ACT-R command to a text-object"
  `(let ((*command-trace* (make-string-output-stream)))
     ,command
     (s-value ,text-object :string
              (get-output-stream-string *command-trace*))
     (opal:update (gv ,text-object :window))))

(defun sgp-p (parameter)
  "Returns ACT-R global parameter value.  Used as a predicate."
  (let ((*command-trace* nil))
    (unwrap (eval (list 'sgp parameter)))))

(defmacro spp-values (parameter)
  "Returns a list of vlues of a parameter for all the rules"
  `(let ((*command-trace* nil)
         (val nil)
         (values nil))
     (dolist (rule (pp) values)
             (setf val (unwrap (eval (list 'spp rule ,parameter))))
             (push-last val values))))

;;; Act-r Vision main window

(defun vision-go ()
  "Opens Act-R control panel window"

  (create-instance 'LOAD-WIN gg:Motif-Load-Gadget
     (:window-title "Load")
     (:message-string "Fetching directory...")
     (:foreground-color *vision-colour*)
     (:dir-input-field-font (opal:get-standard-font :sans-serif nil :small))
     (:dir-input-label-font (opal:get-standard-font :sans-serif :bold :medium))
     (:file-input-field-font (opal:get-standard-font :sans-serif nil :small))
     (:file-input-label-font (opal:get-standard-font :sans-serif :bold :medium))
     (:file-menu-font (opal:get-standard-font :sans-serif :bold :medium))
     (:button-panel-font (opal:get-standard-font :sans-serif :bold :medium))
     (:selection-function (lambda (gadget value)
                            (declare (ignore gadget))
                            (load value))))

  (create-instance 'V-ABOUT gg:Motif-Error-Gadget
     (:window-title "About")
     (:foreground-color *vision-colour*)
     (:string *about-vision*)
     (:beep-p nil))

  (create-instance 'V-MENU gg:Motif-Menubar
     (:items '((" File " nil
                ((" Load file " v-load)
                 (" Quit " v-quit)))
               (" Tools " nil
                ((" Run Control " run-control)
                 (" Traces " open-trace-control)
                 (" Goal Focus " goal-focus-watch)
                 (" Goal Stack " goal-stack-watch)
                 (" Chunk Watch " add-chunk-watch)
                 (" Rule Watch " add-rule-watch)
                 ))
               (" Help " nil
                ((" About " v-about)))))
     (:foreground-color *vision-colour*)
     (:title-font (opal:get-standard-font :sans-serif :bold :medium))
     (:item-font (opal:get-standard-font :sans-serif :bold :small)) )

  (create-instance 'VISION-MAIN inter:interactor-window
     (:left 65)
     (:width  (o-formula (gv V-MENU :width)))
     (:height (o-formula (gv V-MENU :height)))
     (:title "Act-r Vision")
     (:background-color *vision-colour*))

  (s-value VISION-MAIN :aggregate (create-instance 'V-AGG opal:aggregate))

  (opal:add-components V-AGG V-MENU)

  (opal:update VISION-MAIN))

;;; Menubar functions

(defun v-load (gadget menu-item submenu-item)
  "Load any Lisp source or compiled file, or Act-r model"
  (declare (ignore menu-item submenu-item))
  (opal:update (gv gadget :window))
  (gg:display-load-gadget-and-wait LOAD-WIN))

(defun v-about (gadget menu-item submenu-item)
  "Display About"
  (declare (ignore menu-item submenu-item))
  (opal:update (gv gadget :parent :window))
  (gg:Display-Error V-ABOUT))

(defun v-quit (gadget menu-item submenu-item)
  "Quit Act-r Vision"
  (declare (ignore gadget menu-item submenu-item))
  (opal:destroy VISION-MAIN) ;; This windows has some interactor windows
  (dolist (window (set-difference (windows) *not-my-windows*))
    (opal:destroy window)))

;;; Run control

(defun run-control (gadget menu-item submenu-item)
  "Opens the Run Control tool"

  (declare (ignore gadget menu-item submenu-item))

  (create-instance 'RUN-CONTROL inter:interactor-window
     (:top  (o-formula (- (floor opal:*screen-height* 2)
                          (floor (gvl :height) 2))))
     (:width 57) (:height 100)
     (:title "Run Control")
     (:background-color *vision-colour*))

  (s-value RUN-CONTROL :aggregate (create-instance 'RC-AGG opal:aggregate))

  (create-instance 'RUN-BUTTONS gg:Motif-Text-Button-Panel
     (:items '("Reload" "Step" "Run" "Stop"))
     (:v-spacing 0)
     (:h-align :CENTER)
     (:foreground-color *vision-colour*)
     (:font (opal:get-standard-font :sans-serif :bold :medium))
     (:selection-function
      (lambda (gadget value)
        (declare (ignore gadget))
        (cond
         ((equal value "Reload")(reload))
         ((equal value "Step")  (run 1))
         ((equal value "Run")   (run 2000))
         ((equal value "Stop")  (setf *stop* t))
         (t (format t "value ~S~%" value))))) )

  (opal:add-components RC-AGG RUN-BUTTONS)

  (opal:update RUN-CONTROL) )

;;; Trace control

(defun init-trace ()
  "Gets and returns the current trace settings"
  (let ((trace-list nil))
    (when (sgp-p :emt) (push "Exact Matching" trace-list))
    (when (sgp-p :pmt) (push "Partial Matching" trace-list))
    (when (sgp-p :pct) (push "Production Compilation" trace-list))
    (when (sgp-p :act) (push "Activation" trace-list))
    (when (sgp-p :crt) (push "Conflict Resolution" trace-list))
    (when (sgp-p :cst) (push "Conflict Set" trace-list))
    (when (sgp-p :mt)  (push "Matches" trace-list))
    (when (sgp-p :pt)  (push "Production" trace-list))
    (when (sgp-p :ct)  (push "Cycle" trace-list))
    (when (sgp-p :lt)  (push "Latency" trace-list))
    (when (sgp-p :ot)  (push "Output" trace-list))
    (when (sgp-p :dmt) (push "Declarative Memory" trace-list))
    (when (sgp-p :gt)  (push "Goal" trace-list))
    (when (sgp-p :v)   (push "Verbose" trace-list))
    trace-list) )

;;; Set trace

(defun set-trace (gadget value)
  "Sets Act-R traces"
  (declare (ignore gadget))
  (if (member "Exact Matching" value :test #'equal)
      (sgp :emt t) (sgp :emt nil))
  (if (member "Partial Matching" value :test #'equal)
      (sgp :pmt t) (sgp :pmt nil))
  (if (member "Production Compilation" value :test #'equal)
      (sgp :pct t) (sgp :pct nil))
  (if (member "Activation" value :test #'equal)
      (sgp :act t) (sgp :act nil))
  (if (member "Conflict Resolution" value :test #'equal)
      (sgp :crt t) (sgp :crt nil))
  (if (member "Conflict Set" value :test #'equal)
      (sgp :cst t) (sgp :cst nil))
  (if (member "Matches" value :test #'equal)
      (sgp :mt t) (sgp :mt nil))
  (if (member "Production" value :test #'equal)
      (sgp :pt t) (sgp :pt nil))
  (if (member "Cycle" value :test #'equal)
      (sgp :ct t) (sgp :ct nil))
  (if (member "Latency" value :test #'equal)
      (sgp :lt t) (sgp :lt nil))
  (if (member "Output" value :test #'equal)
      (sgp :ot t) (sgp :ot nil))
  (if (member "Declarative Memory" value :test #'equal)
      (sgp :dmt t) (sgp :dmt nil))
  (if (member "Goal" value :test #'equal)
      (sgp :gt t) (sgp :gt nil))
  (if (member "Verbose" value :test #'equal)
      (sgp :v t) (sgp :v nil)))

(defun trace-control ()
  "Trace Control tool"

  (create-instance 'TRACE-CONTROL-WIN inter:interactor-window
      (:top 100) (:width 140) (:height 250)
      (:title "Trace Control")
      (:background-color *vision-colour*))

  (s-value TRACE-CONTROL-WIN :aggregate
           (create-instance 'TC-AGG opal:aggregate))

  (create-instance 'TRACES gg:Motif-Check-Button-Panel
     (:v-spacing 0)
     (:foreground-color *vision-colour*)
     (:font (opal:get-standard-font :sans-serif nil :small))
     (:items '("Exact Matching"
               "Partial Matching"
               "Production Compilation"
               "Activation"
               "Conflict Resolution"
               "Conflict Set"
               "Matches"
               "Production"
               "Cycle"
               "Latency"
               "Output"
               "Declarative Memory"
               "Goal"
               "Verbose"))
     (:selection-function 'set-trace) )

  (opal:add-components TC-AGG TRACES)

  (opal:update TRACE-CONTROL-WIN))

(defun open-trace-control (gadget menu-item submenu-item)
  "Opens the Trace Control tool"
  (declare (ignore gadget menu-item submenu-item))
  (trace-control)
  (s-value TRACES :value (init-trace))
  (opal:update TRACE-CONTROL-WIN))

;;; Goal Focus watch

(defun goal-focus-watch (gadget menu-item submenu-item)
  "Opens the Goal Focus tool"

  (declare (ignore gadget menu-item submenu-item))
  (setf *goal-focus-watch-p* t)
    
  (create-instance 'GOAL-FOCUS-WIN inter:interactor-window
     (:top 0) (:left 450) (:height 280) (:width 180)
     (:title "Goal Focus")
     (:background-color *vision-colour*))

  (s-value GOAL-FOCUS-WIN :aggregate (create-instance 'GF-AGG opal:aggregate))

  (create-instance 'GF-WATCH opal:Text (:top 10)
     (:font (opal:get-standard-font :sans-serif nil :medium)))

  (create-instance 'GF-CLOSE gg:Motif-Text-Button
     (:left (o-formula (- (gvl :parent :window :width) (gvl :width))))
     (:string "Close")
     (:toggle-p NIL)
     (:foreground-color *vision-colour*)
     (:font (opal:get-standard-font :sans-serif :bold :medium))
     (:selection-function
      (lambda (gadget value)
        (declare (ignore value))
        (setf *goal-focus-watch-p* nil)
        (opal:destroy (gv gadget :parent :window)))))

  (opal:add-components GF-AGG GF-WATCH GF-CLOSE)

  (opal:update GOAL-FOCUS-WIN) )

;;; Goal Stack watch

(defun goal-stack-watch (gadget menu-item submenu-item)
  "Opens the Goal Stack tool"

  (declare (ignore gadget menu-item submenu-item))
  (setf *goal-stack-watch-p* t)
    
  (create-instance 'GOAL-STACK-WIN inter:interactor-window
     (:top 0) (:left 210) (:height 120) (:width 220)
     (:title "Goal Stack")
     (:background-color *vision-colour*))

  (s-value GOAL-STACK-WIN :aggregate (create-instance 'GS-AGG opal:aggregate))

  (create-instance 'GS-WATCH opal:Text (:top 10)
     (:font (opal:get-standard-font :sans-serif nil :medium)))

  (create-instance 'GS-CLOSE gg:Motif-Text-Button
     (:left (o-formula (- (gvl :parent :window :width) (gvl :width))))
     (:string "Close")
     (:toggle-p NIL)
     (:foreground-color *vision-colour*)
     (:font (opal:get-standard-font :sans-serif :bold :medium))
     (:selection-function
      (lambda (gadget value)
        (declare (ignore value))
        (setf *goal-stack-watch-p* nil)
        (opal:destroy (gv gadget :parent :window)))))

  (opal:add-components GS-AGG GS-WATCH GS-CLOSE)

  (opal:update GOAL-STACK-WIN) )

;;; Query string

(defun ask-name (&optional string)
  "Opens a dialog to enter a string"

  (create-instance 'ASK-NAME-WIN inter:interactor-window
     (:width  200)
     (:height 40)
     (:left (o-formula (- (floor opal:*screen-width* 2)
                          (floor (gvl :width)  2))))
     (:top  (o-formula (- (floor opal:*screen-height* 2)
                          (floor (gvl :height) 2))))
     (:title  (if string string "Name"))
     (:background-color *vision-colour*))

  (s-value ASK-NAME-WIN :aggregate (create-instance 'ASK-AGG opal:aggregate))

  (create-instance 'NAME-BOX gg:Motif-Scrolling-Labeled-Box
     (:left (o-formula (opal:gv-center-x-is-center-of ASK-NAME-WIN)))
     (:top  (o-formula (opal:gv-center-y-is-center-of ASK-NAME-WIN)))
     (:foreground-color *vision-colour*)
     (:label-font (opal:get-standard-font :sans-serif :bold :medium))
     (:label-string "Name:")
     (:width (o-formula (- (gv ASK-NAME-WIN :width) 20)))
     (:value "type here")
     (:selection-function (lambda (gadget value)
                            (opal:destroy (gv gadget :parent :window))
                            (inter:interaction-complete
                             (gg:careful-read-from-string value)))))

  (opal:add-components ASK-AGG NAME-BOX)

  (opal:update ASK-NAME-WIN) )

;;; Chunk watch

(defun add-chunk-watch (gadget menu-item submenu-item)
  "Opens the Chunk Watch tool"

  (declare (ignore gadget menu-item submenu-item))

  (let* ((chunk (inter:wait-interaction-complete (ask-name "Chunk name")))
         (chunk-name (symbol-name chunk))
         (win (intern (concatenate 'string chunk-name "-WIN")))
         (agg (intern (concatenate 'string chunk-name "-AGG")))
         (watch (intern (concatenate 'string chunk-name "-WATCH")))
         (close (intern (concatenate 'string chunk-name "-CLOSE"))) )

    (push (cons chunk watch) *chunk-watch-list*)
    
    (create-instance win inter:interactor-window
       (:top 30) (:height 280) (:width 180)
       (:title chunk-name)
       (:background-color *vision-colour*))

    (s-value (eval win) :aggregate (create-instance agg opal:aggregate))

    (create-instance watch opal:Text (:top 10)
     (:font (opal:get-standard-font :sans-serif nil :medium)))

    (create-instance close gg:Motif-Text-Button
       (:left (o-formula (- (gvl :parent :window :width) (gvl :width))))
       (:string "Close")
       (:toggle-p NIL)
       (:foreground-color *vision-colour*)
       (:font (opal:get-standard-font :sans-serif :bold :medium))
       (:selection-function
        (lambda (gadget value)
          (declare (ignore value))
          (setf *chunk-watch-list*
                (delete (cons chunk watch) *chunk-watch-list* :test #'equal))
          (opal:destroy (gv gadget :parent :window)))))

    (opal:add-components (eval agg) (eval watch) (eval close))

    (opal:update (eval win))) )

;;; Rule watch

(defun add-rule-watch (gadget menu-item submenu-item)
  "Opens the Rule Watch tool"

  (declare (ignore gadget menu-item submenu-item))

  (let* ((rule (inter:wait-interaction-complete (ask-name "Rule name")))
         (rule-name (symbol-name rule))
         (win (intern (concatenate 'string rule-name "-WIN")))
         (agg (intern (concatenate 'string rule-name "-AGG")))
         (watch (intern (concatenate 'string rule-name "-WATCH")))
         (close (intern (concatenate 'string rule-name "-CLOSE"))) )

    (push (cons rule watch) *rule-watch-list*)
    
    (create-instance win inter:interactor-window
       (:top 30) (:height 300) (:width 190)
       (:title rule-name)
       (:background-color *vision-colour*))

    (s-value (eval win) :aggregate (create-instance agg opal:aggregate))

    (create-instance watch opal:Text (:top 10)
     (:font (opal:get-standard-font :sans-serif nil :medium)))

    (create-instance close gg:Motif-Text-Button
       (:left (o-formula (- (gvl :parent :window :width) (gvl :width))))
       (:string "Close")
       (:toggle-p NIL)
       (:foreground-color *vision-colour*)
       (:font (opal:get-standard-font :sans-serif :bold :medium))
       (:selection-function
        (lambda (gadget value)
          (declare (ignore value))
          (setf *rule-watch-list*
                (delete (cons rule watch) *rule-watch-list* :test #'equal))
          (opal:destroy (gv gadget :parent :window)))))

    (opal:add-components (eval agg) (eval watch) (eval close))

    (opal:update (eval win))) )

;;;
;;; Hooking to ACT-R
;;;

(defun setup-hook-functions ()
  "Config hook functions"
  (setf *cycle-hook-fn* 'cycle-hook-fn))

(defun cycle-hook-fn (&optional instantiation)
  "Call this function after each cycle with the instantiation fired."
  (declare (ignore instantiation))
  (dolist (chunk-watch *chunk-watch-list*)
          (watch-command (eval (list 'dm (car chunk-watch)))
                         (eval (cdr chunk-watch))))
  (dolist (rule-watch *rule-watch-list*)
          (watch-command (eval (list 'spp (car rule-watch)))
                         (eval (cdr rule-watch))))
  (when *goal-focus-watch-p* (watch-command (goal-focus) GF-WATCH))
  (when *goal-stack-watch-p* (watch-command (goal-stack) GS-WATCH)) )

(setf *init-hook-fn* 'setup-hook-functions)

;;;
;;; Launch Vision
;;;

(vision-go)
